   import java.awt.Color;
   import java.util.ArrayList;
   import java.util.Random;
   public class Fish implements Locatable
   {
      private static int nextAvailableID = 1;
      private Environment theEnv;
      private int myId;          
      private Location myLoc;    
      private Direction myDir;   
      private Color myColor;     
      private double probOfBreeding; 
      private double probOfDying;    
      public Fish(Environment env, Location loc)
      {
         initialize(env, loc, env.randomDirection(), randomColor());
      }
      public Fish(Environment env, Location loc, Direction dir)
      {
         initialize(env, loc, dir, randomColor());
      }
      public Fish(Environment env, Location loc, Direction dir, Color col)
      {
         initialize(env, loc, dir, col);
      }
      private void initialize(Environment env, Location loc, Direction dir,
                        Color col)
      {
         theEnv = env;
         myId = nextAvailableID;
         nextAvailableID++;
         myLoc = loc;
         myDir = dir;
         myColor = col;
         theEnv.add(this);
         probOfBreeding = 1.0/7.0;
         probOfDying = 1.0/5.0;   
      }
      protected Color randomColor()
      {
      
         Random randNumGen = RandNumGenerator.getInstance();
         return new Color(randNumGen.nextInt(256),   
                         randNumGen.nextInt(256),   
                         randNumGen.nextInt(256));  
      }
      public int id()
      {
         return myId;
      }
      public Environment environment()
      {
         return theEnv;
      }
      public Color color()
      {
         return myColor;
      }
      public Location location()
      {
         return myLoc;
      }
      public Direction direction()
      {
         return myDir;
      }
      public boolean isInEnv()
      {
         return environment().objectAt(location()) == this;
      }
      public String toString()
      {
         return id() + location().toString() + direction().toString();
      }
      public void act()
      {
         if(!isInEnv())
            return;
      
         if(!breed())
            move();
      
         Random randNumGen = RandNumGenerator.getInstance();
         if(randNumGen.nextDouble() < probOfDying)
            die();
      }
      protected boolean breed()
      {
         Random randNumGen = RandNumGenerator.getInstance();
         if(randNumGen.nextDouble() >= probOfBreeding)
            return false;
      
         ArrayList emptyNbrs = emptyNeighbors();
         Debug.print("Fish " + toString() + " attempting to breed.  ");
         Debug.println("Has neighboring locations: " + emptyNbrs.toString());
      
         if(emptyNbrs.size() == 0)
         {
            Debug.println("  Did not breed.");
            return false;
         }
      
         for(int index = 0; index < emptyNbrs.size(); index++)
         {
            Location loc = (Location) emptyNbrs.get(index);
            generateChild(loc);
         }
         return true;
      }
      protected void generateChild(Location loc)
      {
         Fish child = new Fish(environment(), loc,
                           environment().randomDirection(), color());
         Debug.println("  New Fish created: " + child.toString());
      }
      protected void move()
      {
         Debug.print("Fish " + toString() + " attempting to move.  ");
         Location nextLoc = nextLocation();
      
         if(!nextLoc.equals(location()))
         {
            Location oldLoc = location();
            changeLocation(nextLoc);
         
            Direction newDir = environment().getDirection(oldLoc, nextLoc);
            changeDirection(newDir);
            Debug.println("  Moves to " + location() + direction());
         }
         else
            Debug.println("  Does not move.");
      }
      protected Location nextLocation()
      {
         ArrayList emptyNbrs = emptyNeighbors();
      
         Direction oppositeDir = direction().reverse();
         Location locationBehind = environment().getNeighbor(location(),
                           oppositeDir);
         emptyNbrs.remove(locationBehind);
         Debug.print("Possible new locations are: " + emptyNbrs.toString());
      
         if(emptyNbrs.size() == 0)
            return location();
      
         Random randNumGen = RandNumGenerator.getInstance();
         int randNum = randNumGen.nextInt(emptyNbrs.size());
         return (Location)emptyNbrs.get(randNum);
      }
      protected ArrayList emptyNeighbors()
      {
         ArrayList nbrs = environment().neighborsOf(location());
      
         ArrayList emptyNbrs = new ArrayList();
         for(int index = 0; index < nbrs.size(); index++)
         {
            Location loc = (Location)nbrs.get(index);
            if(environment().isEmpty(loc))
               emptyNbrs.add(loc);
         }
         return emptyNbrs;
      }
      protected void changeLocation(Location newLoc)
      {
         Location oldLoc = location();
         myLoc = newLoc;
         environment().recordMove(this, oldLoc);
      }
      protected void changeDirection(Direction newDir)
      {
         myDir = newDir;
      }
      protected void die()
      {
         Debug.println(toString() + " about to die.");
         environment().remove(this);
      }
   }