   import java.util.*;
   public class UnboundedEnv extends SquareEnvironment
   {
      private TreeMap objectList = new TreeMap();
      public int numRows() {
         return -1;
      }
      public int numCols() {
         return -1;
      }
      public boolean isValid(Location loc) {
         return loc != null;
      }
      public int numObjects() {
         return objectList.size();
      }
      public Locatable[] allObjects() {
         Locatable[] objectArray = new Locatable[objectList.size()];
         Iterator it = objectList.keySet().iterator();
         for(int index = 0; index < objectList.size(); index++)
         {
            objectArray[index] = (Locatable)objectList.get(it.next());
         }
         return objectArray;
      }
      public boolean isEmpty(Location loc) {
         return (objectAt(loc) == null);
      }
      public Locatable objectAt(Location loc) {
         Object x = objectList.get(loc);
         return (x == null) ? null : (Locatable)x;
      }
      public String toString() {
         Locatable[] theObjects = allObjects();
         String s = "Environment has " + numObjects() + " objects: ";
         for (int index = 0; index < theObjects.length; index++)
            s += theObjects[index].toString() + " ";
         return s;
      }
      public void add(Locatable obj) {
         Location loc = obj.location();
         if(!isEmpty(loc))
            throw new IllegalArgumentException("Location " + loc +
               " is not a valid empty location");
         objectList.put(loc, obj);
      }
      public void remove(Locatable obj) {
         Location loc = obj.location();
         if(!objectList.containsKey(loc))
            throw new IllegalArgumentException("Cannot remove " +
               obj + "; not there");
         objectList.remove(loc);
      }
      public void recordMove(Locatable obj, Location oldLoc) {
         Location newLoc = obj.location();
         if(objectList.containsKey(newLoc) && !oldLoc.equals(newLoc) 
         || !objectList.containsKey(oldLoc))
         {
            throw new IllegalArgumentException("Precondition violation "
               + "moving " + obj + " from " + oldLoc);
         }
         objectList.remove(oldLoc);
         objectList.put(newLoc, obj);
      }
   }
